"""This module holds glue code between web servers and rolo. Any ASGI-compliant server can server rolo with
a few patches via ``rolo.serving.asgi``. Moreover, rolo can be served through twisted via
``rolo.serving.twisted``."""
